@@ -1,13 +1,17 @@
 {-# LANGUAGE ForeignFunctionInterface, DeriveDataTypeable #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE CPP #-}
 -- | A port of the direct-sqlite package for dealing directly with
 -- 'PersistValue's.
 module Database.Sqlite  (
                          Connection,
                          Statement,
                          Error(..),
+                         SqliteException(..),
                          StepResult(Row,
                                     Done),
+                         Config(ConfigLogFn),
+                         LogFunction,
                          open,
                          close,
                          prepare,
@@ -23,7 +27,10 @@ module Database.Sqlite  (
                          bind,
                          column,
                          columns,
-                         changes
+                         changes,
+                         mkLogFunction,
+                         freeLogFunction,
+                         config
                         )
     where
 
@@ -31,21 +38,50 @@ import Prelude hiding (error)
 import qualified Prelude as P
 import qualified Prelude
 import qualified Data.ByteString as BS
+import qualified Data.ByteString.Unsafe as BSU
 import qualified Data.ByteString.Internal as BSI
 import Foreign
 import Foreign.C
-import Database.Persist (PersistValue (..), listToJSON, mapToJSON, ZT (ZT))
+import Control.Exception (Exception, throwIO)
+import Database.Persist (PersistValue (..), listToJSON, mapToJSON)
 import Data.Text (Text, pack, unpack)
 import Data.Text.Encoding (encodeUtf8, decodeUtf8With)
 import Data.Text.Encoding.Error (lenientDecode)
 import Data.Monoid (mappend, mconcat)
 import Data.IORef (IORef, newIORef, readIORef, writeIORef)
 import Data.Fixed (Pico)
+import Data.Time (formatTime, UTCTime)
+import Data.Typeable (Typeable)
+
+#if MIN_VERSION_time(1,5,0)
+import Data.Time (defaultTimeLocale)
+#else
+import System.Locale (defaultTimeLocale)
+#endif
 
 data Connection = Connection !(IORef Bool) Connection'
 newtype Connection' = Connection' (Ptr ())
 newtype Statement = Statement (Ptr ())
 
+-- | A custom exception type to make it easier to catch exceptions.
+--
+-- Since 2.1.3
+data SqliteException = SqliteException
+    { seError        :: !Error
+    , seFunctionName :: !Text
+    , seDetails      :: !Text
+    }
+    deriving (Typeable)
+instance Show SqliteException where
+    show (SqliteException error functionName details) = unpack $ mconcat
+        ["SQLite3 returned "
+        , pack $ show error
+        , " while attempting to perform "
+        , functionName
+        , details
+        ]
+instance Exception SqliteException
+
 data Error = ErrorOK
            | ErrorError
            | ErrorInternal
@@ -141,13 +177,11 @@ sqlError maybeConnection functionName error = do
                  details <- errmsg database
                  return $ ": " `mappend` details
                Nothing -> return "."
-  fail $ unpack $ mconcat
-    ["SQLite3 returned "
-    , pack $ show error
-    , " while attempting to perform "
-    , functionName
-    , details
-    ]
+  throwIO SqliteException
+    { seError = error
+    , seFunctionName = functionName
+    , seDetails = details
+    }
 
 foreign import ccall "sqlite3_open"
   openC :: CString -> Ptr (Ptr ()) -> IO Int
@@ -250,16 +284,26 @@ finalize statement = do
     ErrorOK -> return ()
     _ -> return () -- sqlError Nothing "finalize" error
 
+-- Taken from: https://github.com/IreneKnapp/direct-sqlite/blob/master/Database/SQLite3/Direct.hs
+-- | Like 'unsafeUseAsCStringLen', but if the string is empty,
+-- never pass the callback a null pointer.
+unsafeUseAsCStringLenNoNull
+    :: BS.ByteString
+    -> (CString -> Int -> IO a)
+    -> IO a
+unsafeUseAsCStringLenNoNull bs cb
+    | BS.null bs = cb (intPtrToPtr 1) 0
+    | otherwise = BSU.unsafeUseAsCStringLen bs $ \(ptr, len) ->
+        cb ptr (fromIntegral len)
+
 foreign import ccall "sqlite3_bind_blob"
   bindBlobC :: Ptr () -> Int -> Ptr () -> Int -> Ptr () -> IO Int
 bindBlobError :: Statement -> Int -> BS.ByteString -> IO Error
-bindBlobError (Statement statement) parameterIndex byteString = do
-  size <- return $ BS.length byteString
-  BS.useAsCString byteString
-                  (\dataC -> do
-                     error <- bindBlobC statement parameterIndex (castPtr dataC) size
+bindBlobError (Statement statement) parameterIndex byteString =
+  unsafeUseAsCStringLenNoNull byteString $ \dataC size -> do
+    error <- bindBlobC statement parameterIndex (castPtr dataC) size
                                         (intPtrToPtr (-1))
-                     return $ decodeError error)
+    return $ decodeError error
 bindBlob :: Statement -> Int -> BS.ByteString -> IO ()
 bindBlob statement parameterIndex byteString = do
   error <- bindBlobError statement parameterIndex byteString
@@ -322,14 +366,10 @@ bindNull statement parameterIndex = do
 foreign import ccall "sqlite3_bind_text"
   bindTextC :: Ptr () -> Int -> CString -> Int -> Ptr () -> IO Int
 bindTextError :: Statement -> Int -> Text -> IO Error
-bindTextError (Statement statement) parameterIndex text = do
-  let byteString = encodeUtf8 text
-  size <- return $ BS.length byteString
-  BS.useAsCString byteString
-                  (\dataC -> do
-                     error <- bindTextC statement parameterIndex dataC size
-                                        (intPtrToPtr (-1))
-                     return $ decodeError error)
+bindTextError (Statement statement) parameterIndex text =
+  unsafeUseAsCStringLenNoNull (encodeUtf8 text) $ \dataC size -> do
+    error <- bindTextC statement parameterIndex dataC size (intPtrToPtr (-1))
+    return $ decodeError error
 bindText :: Statement -> Int -> Text -> IO ()
 bindText statement parameterIndex text = do
   error <- bindTextError statement parameterIndex text
@@ -351,8 +391,7 @@ bind statement sqlData = do
             PersistNull -> bindNull statement parameterIndex
             PersistDay d -> bindText statement parameterIndex $ pack $ show d
             PersistTimeOfDay d -> bindText statement parameterIndex $ pack $ show d
-            PersistUTCTime d -> bindText statement parameterIndex $ pack $ show d
-            PersistZonedTime (ZT d) -> bindText statement parameterIndex $ pack $ show d
+            PersistUTCTime d -> bindText statement parameterIndex $ pack $ format8601 d
             PersistList l -> bindText statement parameterIndex $ listToJSON l
             PersistMap m -> bindText statement parameterIndex $ mapToJSON m
             PersistDbSpecific s -> bindText statement parameterIndex $ decodeUtf8With lenientDecode s
@@ -361,6 +400,9 @@ bind statement sqlData = do
        $ zip [1..] sqlData
   return ()
 
+format8601 :: UTCTime -> String
+format8601 = formatTime defaultTimeLocale "%FT%T%Q"
+
 foreign import ccall "sqlite3_column_type"
   columnTypeC :: Ptr () -> Int -> IO Int
 columnType :: Statement -> Int -> IO ColumnType
@@ -436,3 +478,53 @@ foreign import ccall "sqlite3_changes"
 
 changes :: Connection -> IO Int64
 changes (Connection _ c) = fmap fromIntegral $ changesC c
+
+-- | Log function callback. Arguments are error code and log message.
+--
+-- Since 2.1.4
+type RawLogFunction = Ptr () -> Int -> CString -> IO ()
+
+foreign import ccall "wrapper"
+  mkRawLogFunction :: RawLogFunction -> IO (FunPtr RawLogFunction)
+
+-- |
+-- Since 2.1.4
+newtype LogFunction = LogFunction (FunPtr RawLogFunction)
+
+-- | Wraps a given function to a 'LogFunction' to be further used with 'ConfigLogFn'.
+-- First argument of given function will take error code, second - log message.
+-- Returned value should be released with 'freeLogFunction' when no longer required.
+mkLogFunction :: (Int -> String -> IO ()) -> IO LogFunction
+mkLogFunction fn = fmap LogFunction . mkRawLogFunction $ \_ errCode cmsg -> do
+  msg <- peekCString cmsg
+  fn errCode msg
+
+-- | Releases a native FunPtr for the 'LogFunction'.
+--
+-- Since 2.1.4
+freeLogFunction :: LogFunction -> IO ()
+freeLogFunction (LogFunction fn) = freeHaskellFunPtr fn
+
+-- | Configuration option for SQLite to be used together with the 'config' function.
+--
+-- Since 2.1.4
+data Config
+  -- | A function to be used for logging
+  = ConfigLogFn LogFunction
+
+foreign import ccall "persistent_sqlite_set_log"
+  set_logC :: FunPtr RawLogFunction -> Ptr () -> IO Int
+
+-- | Sets SQLite global configuration parameter. See SQLite documentation for the <https://www.sqlite.org/c3ref/config.html sqlite3_config> function.
+-- In short, this must be called prior to any other SQLite function if you want the call to succeed.
+--
+-- Since 2.1.4
+config :: Config -> IO ()
+config c = case c of
+  ConfigLogFn (LogFunction rawLogFn) -> do
+    e <- fmap decodeError $ set_logC rawLogFn nullPtr
+    case e of
+      ErrorOK -> return ()
+      _ -> sqlError Nothing "sqlite3_config" e
+
+
